@@ -9,23 +9,23 @@ use List::Util   qw(first);
 
 our $METHOD_DELIM = '_';
 
-use constant ATTR_RW        => 'rw';
-use constant ATTR_RO        => 'ro';
+use constant ATTR_PUBLIC    => 'pu';
 use constant ATTR_PROTECTED => 'pr';
 use constant ATTR_PRIVATE   => 'pv';
-use constant ATTR_META      => 'mt';
+
+our %owner;
 
 =head1 NAME
 
-Class::Attribute - Another way to define class attributes!!!
+Class::Attribute - A fast and light weight alternative for defining class attributes.
 
 =head1 VERSION
 
-Version 0.023
+Version 0.024
 
 =cut
 
-our $VERSION = '0.023';
+our $VERSION = '0.024';
 
 require XSLoader;
 XSLoader::load('Class::Attribute', $VERSION);
@@ -41,19 +41,17 @@ found on CPAN; Class::Accessor, Class::InsideOut, Object::InsideOut, Moose and M
 Class::Attribute,
 
 1. Allows definition of attributes of the following type/permissions.
+
     a. public
     b. private
     c. protected
     d. readonly
-    e. meta
 
 2. Has a very simple and extensible attribute validation mechanism using
    regular expressions.
 
 3. Can set default values for attributes.
 
-and thats it.
-
 
     # example
 
@@ -65,7 +63,7 @@ and thats it.
     # pushes itself into @MyUser::ISA.
     use Class::Attribute;
 
-    attribute 'id',      is_int, protected, default => 1;
+    attribute 'id',      is_int, protected, public_accessor, default => 1;
     attribute 'name',    is_string(10, 40, 'a name of 10 - 40 chars');
     attribute 'sex',     is_in( ['M', 'F'], 'sex (M/F)' ), private;
     attribute 'age',     is_int, default => 18;
@@ -123,8 +121,7 @@ and thats it.
 
 =head1 EXPORT
 
-attribute, public, private, protected, readonly, meta, is_int, is_float, is_email,
-is_boolean, is_string, is_date, is_datetime, is_in
+attribute, public, private, protected, readonly, public_accessor, protected_accessor, private_accessor, public_mutator, protected_mutator, private_mutator, is_int, is_float, is_email, is_boolean, is_string, is_date, is_datetime, is_in
 
 =head1 FUNCTIONS
 
@@ -143,7 +140,9 @@ sub import {
     no strict 'refs';
     for my $func qw(
         attribute
-        readonly public private protected meta
+        readonly public private protected
+        public_accessor protected_accessor private_accessor
+        public_mutator protected_mutator private_mutator
         is_int is_float
         is_email is_boolean
         is_string is_in
@@ -153,7 +152,7 @@ sub import {
     }
 
     # avoid repeated inheritence and re-using slots.
-    if (!first { $_ eq __PACKAGE__ } @{"$class:\:ISA"}) {
+    if (!$class->can('_init_class_config')) {
         push @{"$class:\:ISA"}, __PACKAGE__;
         $class->_init_class_config;
     }
@@ -185,11 +184,11 @@ sub _new {
 
     my $self = bless [], $class;
     my $defaults = $class->_defaults;
+    my $slots = $class->_slots;
     for my $attr (keys %$defaults) {
         my $value = $defaults->{$attr};
         $value = &$value() if ref $value eq 'CODE';
-        no strict 'refs'; my $mutator = "set_$attr";
-        $self->$mutator($value);
+        $self->[$slots->{$attr}] = $value;
     }
     return $self;
 }
@@ -207,30 +206,17 @@ and setting up appropriate methods in the caller namespace.
 
 The possible attribute definitions are explained below.
 
-1. Attribute Permission - public (default), private, protected, readonly or meta.
+1. Attribute Permission. This can be defined for accessors and mutators as I<public (default), private, protected or readonly> or for finer grain control separately as I<public_accessor, public_mutator, protected_accessor, protected_mutator, private_accessor or private_mutator>.
 
  # e.g.  attribute 'id1', public;
- # e.g.  attribute 'id2', private;
+ # e.g.  attribute 'id2', private, protected_accessor;
  # e.g.  attribute 'id3', protected;
  # e.g.  attribute 'id4', readonly;
- # e.g.  attribute 'id4', meta;
 
 
 I<CAVEATS>
 
-The following matrix illustrates the types of accessors and mutators that
-will be created given the attribute type/permission.
-
-            |    accessor    |    mutator
- -----------|----------------|--------------
- public     |    public      |    pubic
- private    |    public      |    private
- protected  |    public      |    protected
- readonly   |    public      |    N/A
- meta       |    protected   |    protected
-
-The readonly attribute definition will result in no accessor being created.
-The meta attribute has only a private accessor and mutator.
+The readonly attribute definition will result in a public accessor and private mutator.
 
 2. Attribute ISA - Checks the value in the validator if it is an instance of a particular class.
 
@@ -262,7 +248,12 @@ sub attribute {
     my $name  = shift @_;
     my %attrs = @_;
 
-    my $perm      = delete $attrs{perms} || ATTR_RW;
+    my ($perm_read, $perm_write) = delete @attrs{qw(perm_read perm_write)};
+    if (!$perm_read && !$perm_write) {
+        $perm_read  = ATTR_PUBLIC;
+        $perm_write = ATTR_PUBLIC;
+    }
+
     my $validates = delete $attrs{like};
     my $default   = delete $attrs{default};
     my $isa       = delete $attrs{isa};
@@ -283,31 +274,36 @@ sub attribute {
     $class->_slots->{$name} = $slot;
 
     my $package = __PACKAGE__;
-    # meta types default to private accessors.
+    if ($perm_read) {
+        if ($perm_read eq ATTR_PUBLIC) {
+            __PACKAGE__->_make_accessor("$class\::get$METHOD_DELIM$name", $slot);
+        }
+        elsif ($perm_read eq ATTR_PROTECTED) {
+            __PACKAGE__->_make_protected_accessor("$class\::get$METHOD_DELIM$name", $slot);
+        }
+        else {
+            $owner{$class}{$slot} = $name;
+            $class->_method_owner->[$slot][0] = "$class\::get$METHOD_DELIM$name";
+            __PACKAGE__->_make_private_accessor("$class\::get$METHOD_DELIM$name", $slot);
+        }
+    }
 
-    if ($perm eq ATTR_META) {
-        __PACKAGE__->_make_protected_accessor("$class\::get$METHOD_DELIM$name", $slot);
-        __PACKAGE__->_make_protected_mutator ("$class\::set$METHOD_DELIM$name", $slot);
-    }
-    elsif ($perm eq ATTR_PROTECTED) {
-        __PACKAGE__->_make_accessor("$class\::get$METHOD_DELIM$name", $slot);
-        __PACKAGE__->_make_protected_mutator ("$class\::set$METHOD_DELIM$name", $slot);
-    }
-    elsif ($perm eq ATTR_PRIVATE) {
-        __PACKAGE__->_make_accessor("$class\::get$METHOD_DELIM$name", $slot);
-        __PACKAGE__->_make_private_mutator ("$class\::set$METHOD_DELIM$name", $slot);
-    }
-    elsif ($perm eq ATTR_RO) {
-        __PACKAGE__->_make_accessor("$class\::get$METHOD_DELIM$name", $slot);
-    }
-    elsif ($perm eq ATTR_RW) {
-        __PACKAGE__->_make_accessor("$class\::get$METHOD_DELIM$name", $slot);
-        __PACKAGE__->_make_mutator ("$class\::set$METHOD_DELIM$name", $slot);
+    if ($perm_write) {
+        if ($perm_write eq ATTR_PUBLIC) {
+            __PACKAGE__->_make_mutator ("$class\::set$METHOD_DELIM$name", $slot);
+        }
+        elsif ($perm_write eq ATTR_PROTECTED) {
+            __PACKAGE__->_make_protected_mutator ("$class\::set$METHOD_DELIM$name", $slot);
+        }
+        else {
+            $owner{$class}{$slot} = $name;
+            $class->_method_owner->[$slot][1] = "$class\::set$METHOD_DELIM$name";
+            __PACKAGE__->_make_private_mutator ("$class\::set$METHOD_DELIM$name", $slot);
+        }
     }
 
     if ($predicate) {
-        *{"$class\::$predicate"} = eval "sub { !!shift->get$METHOD_DELIM$name }";
-        croak $@ if $@;
+        __PACKAGE__->_make_predicate("$class\::$predicate", $slot);
     }
 
     if ($isa) {
@@ -318,98 +314,6 @@ sub attribute {
     }
 }
 
-# TODO XSify this.
-sub _make_protected_accessor {
-    my ($class, $method, $slot) = @_;
-
-    (my $realmethod = rand()) =~ s/\./_/;
-    $realmethod = $method . $realmethod;
-    $class->_make_accessor($realmethod, $slot);
-
-    no strict 'refs';
-    *$method = sub {
-        my ($self) = @_;
-        my ($caller, $subroutine) = (caller(0))[0, 3];
-        (my $class) = $subroutine =~ m/^(.*)::/;
-
-        if ($caller eq $class or $caller eq __PACKAGE__ or $caller->isa($class)) {
-            goto &$realmethod;
-        }
-        else {
-            croak("Called a protected accessor $method");
-        }
-    };
-}
-
-# TODO XSify this.
-sub _make_private_accessor {
-    my ($class, $method, $slot) = @_;
-
-    (my $realmethod = rand()) =~ s/\./_/;
-    $realmethod = $method . $realmethod;
-    $class->_make_accessor($realmethod, $slot);
-
-    no strict 'refs';
-    *$method = sub {
-        my ($self) = @_;
-        my ($caller, $subroutine) = (caller(0))[0, 3];
-        (my $class) = $subroutine =~ m/^(.*)::/;
-
-        if ($caller eq $class or $caller eq __PACKAGE__) {
-            goto &$realmethod;
-        }
-        else {
-            croak("Called a private accessor $method");
-        }
-    };
-}
-
-# TODO XSify this.
-sub _make_protected_mutator {
-    my ($class, $method, $slot) = @_;
-
-    (my $realmethod = rand()) =~ s/\./_/;
-    $realmethod = $method . $realmethod;
-    $class->_make_mutator($realmethod, $slot);
-
-    no strict 'refs';
-    *$method = sub {
-        my ($self, $value) = @_;
-        my ($caller, $subroutine) = (caller(0))[0, 3];
-        (my $class) = $subroutine =~ m/^(.*)::/;
-
-        if ($caller eq $class or $caller eq __PACKAGE__ or $caller->isa($class)) {
-            goto &$realmethod;
-        }
-        else {
-            croak("Called a protected mutator $method");
-        }
-    };
-}
-
-# TODO XSify this.
-sub _make_private_mutator {
-    my ($class, $method, $slot) = @_;
-
-    (my $realmethod = rand()) =~ s/\./_/;
-    $realmethod = $method . $realmethod;
-    $class->_make_mutator($realmethod, $slot);
-
-    no strict 'refs';
-    *$method = sub {
-        my ($self, $value) = @_;
-        my ($caller, $subroutine) = (caller(0))[0, 3];
-        (my $class) = $subroutine =~ m/^(.*)::/;
-
-        if ($caller eq $class or $caller eq __PACKAGE__) {
-            goto &$realmethod;
-        }
-        else {
-            croak("Called a private mutator $method");
-        }
-    };
-}
-
 =head2 validate
 
 validates the attributes against any specified validators and returns
@@ -444,11 +348,12 @@ sub validate {
 sub _init_class_config {
     croak 'Cannot access private method _init_class_config' unless caller eq __PACKAGE__;
     my ($class) = @_;
-    _define_config_attribute($class, $_) for qw(_slots _slot_count _defaults _validator);
+    _define_hash_variable($class, $_) for qw(_slots _slot_count _defaults _validator);
+    _define_array_variable($class, $_) for qw(_method_owner);
 }
 
-sub _define_config_attribute {
-    croak 'Cannot access private method _define_config_attribute' unless caller eq __PACKAGE__;
+sub _define_hash_variable {
+    croak 'Cannot access private method _define_hash_variable' unless caller eq __PACKAGE__;
     my ($class, $var) = @_;
 
     no strict 'refs';
@@ -459,6 +364,27 @@ sub _define_config_attribute {
     return ${"$class\::$var"};
 }
 
+sub _define_array_variable {
+    croak 'Cannot access private method _define_array_variable' unless caller eq __PACKAGE__;
+    my ($class, $var) = @_;
+
+    no strict 'refs';
+    ${"$class\::$var"} ||= [];
+    *{"$class\::$var"} = sub {
+        no strict 'refs';
+        if (@_ == 2) {
+            return ${"$class\::$var"}->[$_[1]];
+        }
+        if (@_ == 3) {
+            return ${"$class\::$var"}->[$_[1]]->[$_[2]];
+        }
+        else {
+            return ${"$class\::$var"};
+        }
+    };
+    return ${"$class\::$var"};
+}
+
 =head2 _get_all_attributes
 
 This is a private method that returns a hash of attribute names and their values.
@@ -474,14 +400,20 @@ sub _get_all_attributes {
     return map { $_ => $self->[$slotmap{$_}] } keys %slotmap;
 }
 
-
 # some constants and default validators.
+use constant readonly  =>  perm_read => ATTR_PUBLIC,    perm_write => ATTR_PRIVATE;
+use constant public    =>  perm_read => ATTR_PUBLIC,    perm_write => ATTR_PUBLIC;
+use constant protected =>  perm_read => ATTR_PROTECTED, perm_write => ATTR_PROTECTED;
+use constant private   =>  perm_read => ATTR_PRIVATE,   perm_write => ATTR_PRIVATE;
+
+use constant public_accessor    => perm_read => ATTR_PUBLIC;
+use constant protected_accessor => perm_read => ATTR_PROTECTED;
+use constant private_accessor   => perm_read => ATTR_PRIVATE;
+
+use constant public_mutator     => perm_write => ATTR_PUBLIC;
+use constant protected_mutator  => perm_write => ATTR_PROTECTED;
+use constant private_mutator    => perm_write => ATTR_PRIVATE;
 
-use constant readonly  =>  'perms' , ATTR_RO;
-use constant public    =>  'perms' , ATTR_RW;
-use constant protected =>  'perms' , ATTR_PROTECTED;
-use constant private   =>  'perms' , ATTR_PRIVATE;
-use constant meta      =>  'perms' , ATTR_META;
 
 # numeric
 use constant INT       => qr/^\d+$/;
